
class TreeNode:
    """Node class for binary tree

    Attributes:
        key: key of the node
        val: value of the node
        left_child: TreeNode object. The left child of the current node
        right_child: TreeNode object. The right child of the current node
        parent: TreeNode object. The parents of the current node
    """

    def __init__(self, key, val, left_child=None, right_child=None, parent=None):
        """Initialises the TreeNode object

            Arguments:
                key: key of the root node
                val: value of the root node
                left_child: TreeNode object. The left child of the current node (default = None)
                right_child: TreeNode object. The right child of the current node (default = None)
                parent: TreeNode object. The parent of the current node (default = None)
        """
        self.key = key
        self.val = val
        self.left_child = left_child
        self.right_child = right_child
        self.parent = parent

    def get_key(self):
        """Getter method for key"""
        return self.key

    def set_key(self, item):
        """Setter method for key"""
        self.key = item

    def get_val(self):
        """Getter method for val"""
        return self.val

    def set_val(self, item):
        """Setter method for val"""
        self.val = item

    def get_left_child(self):
        """Getter method for the left child"""
        return self.left_child

    def get_right_child(self):
        """Getter method for the right child"""
        return self.right_child

    def is_leaf(self):
        """Check if the current node is a leaf"""
        return not self.left_child and not self.right_child

    def has_left_child_only(self):
        """Check if the current node has a left child only"""
        return self.left_child and not self.right_child

    def has_right_child_only(self):
        """Check if the current node has a right child only"""
        return self.right_child and not self.left_child

    def get_successor(self):
        """Calls find_min() to get the successor node"""
        return self.right_child.find_min()

    def find_min(self):
        """Finds the node with minimum key in the tree"""
        current_node = self
        while current_node.left_child:
            # keep going down the left hand side of the tree to find the min node
            current_node = current_node.left_child
        return current_node

    def splice(self):
        """Splices out a node"""
        if self.right_child:
            # remove a node and attach its right branch to the left side of the parent
            self.parent.left_child = self.right_child
        else:
            # if no right child, remove the left child of the parent
            self.parent.left_child = None

    def __iter__(self):
        if self:
            if self.get_left_child():
                for i in self.left_child:
                    yield i
            yield self.key
            if self.get_right_child():
                for i in self.right_child:
                    yield i


class BinaryTree:
    """Binary tree class

        Attributes:
            root: the root node of the tree (default = None)
            size: the number of elements in the binary tree (default = 0)
    """

    def __init__(self):
        """Initialises the binary tree"""
        self.root = None
        self.size = 0

    def get_size(self):
        """Getter method for size"""
        return self.size

    def insert_node(self, key, val):
        """Inserts a new node into the tree at the correct location"""
        if self.root:
            self._put(key, val, self.root)
            self.size += 1
        else:
            self.root = TreeNode(key, val)
            self.size += 1

    def _put(self, key, val, current_node):
        """Helper method for insert_node"""
        if key < current_node.key:
            if current_node.get_left_child():
                # if current_node has a left child and key < current key, recursively call _put to continue searching
                self._put(key, val, current_node.left_child)
            else:
                # if current node does not have a left child, insert node as left child
                current_node.left_child = TreeNode(key, val, parent=current_node)
        elif key > current_node.key:
            if current_node.get_right_child():
                # if current_node has a right child and key >= current key, recursively call _put to continue searching
                self._put(key, val, current_node.right_child)
            else:
                # if current node does not have a right child, insert node as right child
                current_node.right_child = TreeNode(key, val, parent=current_node)
        else:
            current_node.val = val

    def get_node(self, key):
        """Returns the value of the node with a matching key"""
        if self.root:
            result = self._get(key, self.root)
            if result:
                return result
            else:
                return None
        else:
            return None

    def _get(self, key, current_node):
        """Helper method for get_node"""
        if not current_node:
            return None
        elif current_node.key > key:
            # if target key is smaller than node, recursively search the left hand side of the tree
            return self._get(key, current_node.left_child)
        elif current_node.key < key:
            # if target key is larger than node, recursively search the right hand side of the tree
            return self._get(key, current_node.right_child)
        elif current_node.key == key:
            # if target key equal to node, return node
            return current_node

    def delete(self, key):
        """Deletes a node with matching key"""
        if self.size > 1:
            # get the correct node...
            result = self._get(key, self.root)
            if result:
                # then remove the node, and decrement size
                self._remove_node(result)
                self.size -= 1
            else:
                print('Error! {} not found in Tree'.format(key))
        elif self.get_size() == 1:
            # return the root if it is the only node in the tree
            if self.root.key == key:
                self.root = None
                self.size -= 1
            else:
                print('Error! {} not found in Tree'.format(key))
        else:
            print('Error! {} not found in Tree'.format(key))

    @staticmethod
    def _remove_node(node):
        """Helper method for delete"""
        if node.is_leaf():
            # if the node has no children, just remove the node
            if node.parent.left_child == node:
                node.parent.left_child = None
            elif node.parent.right_child == node:
                node.parent.right_child = None
        elif node.has_left_child_only():
            if node.parent.left_child == node:
                # if node is a left child set the parent left child to the left child branch and delete the node
                node.parent.left_child = node.left_child
            elif node.parent.right_child == node:
                # if node is a right child set the parent right child to the left child branch and delete the node
                node.parent.right_child = node.left_child
        elif node.has_right_child_only():
            if node.parent.left_child == node:
                # if node is a left child set the parent left child to the right child branch and delete the node
                node.parent.left_child = node.right_child
            elif node.parent.right_child == node:
                # if node is a right child set the parent right child to the right child branch and delete the node
                node.parent.right_child = node.right_child
        else:
            # if node has both children, find the successor and splice it in
            succ = node.get_successor()
            succ.splice()
            node.key = succ.key
            node.val = succ.val

    def get_height(self, key):
        """Calculates and returns the height of a node with matching key. Root height = 0"""
        current_node = self._get(key, self.root)
        height_counter = 0
        # find target node and follow up the tree, tracking the height
        while current_node.parent:
            current_node = current_node.parent
            height_counter += 1
        return height_counter

    def __setitem__(self, key, val):
        self.insert_node(key, val)

    def __getitem__(self, key):
        return self.get_node(key)

    def __contains__(self, key):
        if self.get_node(key):
            return True
        else:
            return False

    def __delitem__(self, key):
        self.delete(key)


# if __name__ == '__main__':
#     tree = BinaryTree()
#     key_list = [43, 55, 22, 15, 98, 1, 34, 76]
#     val_list = ['Penguin', 'Ant', 'Bear', 'Shark', 'Eagle', 'Dog', 'Cat', 'Turtle']
#
#     for i, j in zip(key_list, val_list):
#         tree[i] = j
#
#     print(tree[43].get_val())
#     print(43 in tree)



from PyDSL.binary_tree import BinaryTree
from PyDSL.binary_tree import TreeNode


class AVLNode(TreeNode):
    """Node class for AVL (self-balancing) tree. Extends the TreeNode class

        Attributes:
            key: key of the node
            val: value of the node
            left_child: TreeNode object. The left child of the current node
            right_child: TreeNode object. The right child of the current node
            parent: TreeNode object. The parents of the current node
            balance_factor: the balance factor of the node
    """

    def __init__(self, key, val, left_child=None, right_child=None, parent=None):
        """Initialise the AVLNode object

            Arguments:
                key: key of the root node
                val: value of the root node
                left_child: TreeNode object. The left child of the current node (default = None)
                right_child: TreeNode object. The right child of the current node (default = None)
                parent: TreeNode object. The parent of the current node (default = None)
        """
        super().__init__(key, val, left_child, right_child, parent)
        self.balance_factor = 0


class AVLTree(BinaryTree):
    """AVL tree class. Extends BinaryTree

        Attributes:
            root: the root node of the tree (default = None)
            size: the number of elements in the binary tree (default = 0)
    """
    def __init__(self):
        """Initialise the AVLTree object"""
        super().__init__()

    def insert_node(self, key, val):
        """Insert a new node into the tree at the correct location. Overrides insert_node method in BinaryTree"""
        if self.root:
            self._put(key, val, self.root)
            self.size += 1
        else:
            self.root = AVLNode(key, val)
            self.size += 1

    def _put(self, key, val, current_node):
        """Helper method for insert_node. Overrides _put method in BinaryTree"""
        if key < current_node.key:
            if current_node.get_left_child():
                # if current_node has a left child and key < current key, recursively call _put to continue searching
                self._put(key, val, current_node.left_child)
            else:
                # if current node does not have a left child, insert node as left child and update balanace
                current_node.left_child = AVLNode(key, val, parent=current_node)
                self.update_balance(current_node.left_child)
        else:
            if current_node.get_right_child():
                # if current_node has a right child and key >= current key, recursively call _put to continue searching
                self._put(key, val, current_node.right_child)
            else:
                # if current node does not have a right child, insert node as right child and update balance
                current_node.right_child = AVLNode(key, val, parent=current_node)
                self.update_balance(current_node.right_child)

    def update_balance(self, node):
        """Update the balance factor of every node, and rebalance the tree if necessary"""
        if node.balance_factor > 1 or node.balance_factor < -1:
            # call rebalance if the node is out of balance
            self.rebalance(node)
            return
        if node.parent:
            # adjust the parent balance
            if node.parent.left_child == node:
                node.parent.balance_factor += 1
            elif node.parent.right_child == node:
                node.parent.balance_factor -= 1
            # recursively update the parent if out of balance
            if node.parent.balance_factor != 0:
                self.update_balance(node.parent)

    def rotate_left(self, org_root):
        """Implement left rotation of a node"""
        new_root = org_root.right_child
        org_root.right_child = new_root.left_child
        # locate the nodes to rotate
        if new_root.left_child:
            new_root.left_child.parent = org_root
        new_root.parent = org_root.parent
        # if root is rotated, set a new root for the tree
        if not org_root.parent:
            self.root = new_root
        else:
            if org_root.parent.left_child == org_root:
                org_root.parent.left_child = new_root
            elif org_root.parent.right_child == org_root:
                org_root.parent.right_child = new_root
        # perform rotation
        new_root.left_child = org_root
        org_root.parent = new_root
        # update the balance factors of parent and child
        org_root.balance_factor = org_root.balance_factor + 1 - min(new_root.balance_factor, 0)
        new_root.balance_factor = new_root.balance_factor + 1 + max(org_root.balance_factor, 0)

    def rotate_right(self, org_root):
        """Implement right rotation of a node"""
        new_root = org_root.left_child
        org_root.left_child = new_root.right_child
        # locate the nodes to rotate
        if new_root.right_child:
            new_root.right_child.parent = org_root
        new_root.parent = org_root.parent
        # if root is rotated, set a new root for the tree
        if not org_root.parent:
            self.root = new_root
        else:
            if org_root.parent.left_child == org_root:
                org_root.parent.left_child = new_root
            else:
                org_root.parent.right_child = new_root
        # perform rotation
        new_root.right_child = org_root
        org_root.parent = new_root
        # update the balance factors of parent and child
        org_root.balance_factor = org_root.balance_factor - 1 - max(new_root.balance_factor, 0)
        new_root.balance_factor = new_root.balance_factor - 1 + min(org_root.balance_factor, 0)

    def rebalance(self, node):
        """Rebalance the tree"""
        if node.balance_factor < 0:
            if node.right_child.balance_factor > 0:
                # perform two rotations for a "crooked" branch
                self.rotate_right(node.right_child)  # straighten branch
                self.rotate_left(node)
                # perform a single rotation for a straight branch
            else:
                self.rotate_left(node)
        elif node.balance_factor > 0:
            if node.left_child.balance_factor < 0:
                # perform two rotations for a "crooked" branch
                self.rotate_left(node.left_child)  # straighten branch
                self.rotate_right(node)
                # perform a single rotation for a straight branch
            else:
                self.rotate_right(node)


# if __name__ == '__main__':
#     avltree = AVLTree()
#     key_list = [43, 55, 22, 15, 98, 1, 34, 76, 2, 3]
#     val_list = ['Penguin', 'Ant', 'Bear', 'Shark', 'Eagle', 'Dog', 'Cat', 'Turtle', 'Lion', 'Fish']
#
#     for i, j in zip(key_list, val_list):
#         print('Inserting {}'.format(i))
#         avltree[i] = j
#
#     avl_height_list = []
#     for i in key_list:
#         print(i)
#         avl_height_list.append(avltree.get_height(i))
#     for i, j in zip(key_list, avl_height_list):
#         print((i, j))




class HeapNode:
    """HeapNode class. Forms the elements of the MaxHeap and MinHeap classes

        Attributes:
            data: the data contained in the HeapNode. Can be a float/int or a tuple, with the zeroth element of the
            tuple acting as the key
    """

    def __init__(self, data):
        """Initialise the HeapNode object"""
        self.data = data

    def get_data(self):
        """Get the data contained in the HeapNode"""
        return self.data

    def set_data(self, item):
        """Set the data contained in the HeapNode"""
        self.data = item

    def __lt__(self, other):
        """Allows for comparison of HeapNodes. With tuple data, compares the zeroth element of the tuples"""
        if type(self.data) is float or type(self.data) is int:
            if self.data < other.data:
                # compare data attribute directly if data is a float/int
                return True
            else:
                return False
        elif type(self.data) is tuple:
            if self.data[0] < other.data[0]:
                # compare zeroth element of data attribute if data is tuple
                return True
            else:
                return False

    def __gt__(self, other):
        """Allows for comparison of HeapNodes. With tuple data, compares the zeroth element of the tuples"""
        if type(self.data) is float or type(self.data) is int:
            if self.data > other.data:
                # compare data attribute directly if data is a float/int
                return True
            else:
                return False
        elif type(self.data) is tuple:
            if self.data[0] > other.data[0]:
                # compare zeroth element of data attribute if data is tuple
                return True
            else:
                return False

    def __eq__(self, other):
        """Allows for comparison of HeapNodes. With tuple data, compares the zeroth element of the tuples"""
        if type(self.data) is float or type(self.data) is int:
            if self.data == other.data:
                # compare data attribute directly if data is a float/int
                return True
            else:
                return False
        elif type(self.data) is tuple:
            if self.data[0] == other.data[0]:
                # compare zeroth element of data attribute if data is tuple
                return True
            else:
                return False

    def __repr__(self):
        return str(self.data)


class MinHeap:
    """Class for a min heap - a heap that prioritises the smallest item

        Attributes:
            heap_list: the list of HeapNodes. The first element is always an empty node that is ignored by operations
            heap_size: the number of elements in the heap. Starts at 0
    """

    def __init__(self):
        """Initialise the MinHeap"""
        self.heap_list = [HeapNode(None)]
        self.heap_size = 0

    def _percolate_up(self, i):
        """Move an item up the list"""
        while i // 2 > 0:
            # swap elements if i//2th element is bigger than ith element
            if self.heap_list[i // 2] > self.heap_list[i]:
                self.heap_list[i // 2], self.heap_list[i] = self.heap_list[i], self.heap_list[i // 2]
            i = i // 2  # move up the heap_list

    def insert(self, item):
        """Insert an item into the heap"""
        self.heap_list.append(HeapNode(item))
        self.heap_size += 1
        self._percolate_up(self.heap_size)  # place new item into the correct place within heap_list

    def _percolate_down(self, i):
        """Move an item down the list"""
        while i * 2 < self.heap_size:  # keep going until the last layer of the tree is reached
            # compare the left and right children of the ith node. If the left one is smaller...
            if self.heap_list[i * 2] < self.heap_list[i * 2 + 1]:
                # if the ith node is bigger than the selected node, swap the node and its smaller child
                if self.heap_list[i] > self.heap_list[i * 2]:
                    self.heap_list[i], self.heap_list[i * 2] = self.heap_list[i * 2], self.heap_list[i]
                i = i * 2 # move to next level
            # compare the left and right children of the ith node. If the right one is smaller...
            else:
                # if the ith node is bigger than the selected node, swap the node and its smaller child
                if self.heap_list[i] > self.heap_list[i * 2 + 1]:
                    self.heap_list[i], self.heap_list[i * 2 + 1] = self.heap_list[i * 2 + 1], self.heap_list[i]
                i = i * 2 + 1 # move to next level

    def pop(self):
        """Remove and return the HeapNode of highest priority"""
        if self.heap_size > 1:
            # return the first item in the heap_list, replace with the last item and percolate that down to its
            # correct position
            last_item = self.heap_list.pop()
            min_item = self.heap_list[1]
            self.heap_list[1] = last_item
            self.heap_size -= 1
            self._percolate_down(1)
            return min_item
        else:
            # if the heap_list has only one item, return said item
            min_item = self.heap_list.pop()
            self.heap_size -= 1
            return min_item

    def delete_item(self, item_index):
        """Delete a HeapNode in a specific position (index) within heap_list"""
        # locate target node and replace with last node in the heap_list
        target = self.heap_list[item_index]
        last_item = self.heap_list.pop()
        if item_index == self.heap_size:
            self.heap_size -= 1
        else:
            self.heap_list[item_index] = last_item
            self.heap_size -= 1
            self._percolate_down(item_index)  # percolate down to place replacement node in the correct place
        return target

    def build(self, new_list):
        """Construct a MaxHeap given a list of numbers or tuples"""
        for i in new_list:
            self.insert(i)

    def heap_change_key(self, val, new_key):
        """For heaps containing HeapNodes of tuples - change the keys of items with values val"""
        for index, item in enumerate(self.heap_list):
            if index > 0:
                # delete the node with matching val
                if item.get_data()[1] == val:
                    self.delete_item(index)
        # insert a new node object with the new key and same val
        self.insert((new_key, val))

    def __repr__(self):
        return str([node for node in self.heap_list if node.get_data()])

    def __contains__(self, val):
        contained = False
        for node in self.heap_list:
            if type(node.get_data()) is tuple:
                if node.get_data()[1] == val:
                    contained = True
                    break
            else:
                if node.get_data() == val:
                    contained = True
                    break
        return contained

    def __iter__(self):
        return iter(node.get_data() for node in self.heap_list)


class MaxHeap:
    """Class for a max heap - a heap that prioritises the biggest item

        Attributes:
            heap_list: the list of HeapNodes. The first element is always an empty node that is ignored by operations
            heap_size: the number of elements in the heap. Starts at 0
    """

    def __init__(self):
        """Initialise the MaxHeap"""
        self.heap_list = [HeapNode(None)]
        self.heap_size = 0

    def _percolate_up(self, i):
        """Move an item up the list"""
        while i // 2 > 0:
            # swap elements if i//2th element is smaller than ith element
            if self.heap_list[i // 2] < self.heap_list[i]:
                self.heap_list[i // 2], self.heap_list[i] = self.heap_list[i], self.heap_list[i // 2]
            i = i // 2  # move up the heap_list

    def insert(self, item):
        """Insert an item into the heap"""
        self.heap_list.append(HeapNode(item))
        self.heap_size += 1
        self._percolate_up(self.heap_size)  # place new item into the correct place within heap_list

    def _percolate_down(self, i):
        """Move an item down the list"""
        while i * 2 < self.heap_size:
            # keep going until the last layer of the tree is reached
            if self.heap_list[i * 2] > self.heap_list[i * 2 + 1]:
                # compare the left and right children of the ith node. If the left one is bigger...
                if self.heap_list[i] < self.heap_list[i * 2]:
                    # if the ith node is smaller than the selected node, swap the node and its bigger child
                    self.heap_list[i], self.heap_list[i * 2] = self.heap_list[i * 2], self.heap_list[i]
                i = i * 2  # move to next level
            else:
                # compare the left and right children of the ith node. If the right one is bigger...
                if self.heap_list[i] < self.heap_list[i * 2 + 1]:
                    # if the ith node is smaller than the selected node, swap the node and its bigger child
                    self.heap_list[i], self.heap_list[i * 2 + 1] = self.heap_list[i * 2 + 1], self.heap_list[i]
                i = i * 2 + 1  # move to next level

    def pop(self):
        """Remove and return the HeapNode of highest priority"""
        if self.heap_size > 1:
            # return the first item in the heap_list, replace with the last item and percolate that down to its
            # correct position
            last_item = self.heap_list.pop()
            min_item = self.heap_list[1]
            self.heap_list[1] = last_item
            self.heap_size -= 1
            self._percolate_down(1)
            return min_item
        else:
            # if the heap_list has only one item, return said item
            min_item = self.heap_list.pop()
            self.heap_size -= 1
            return min_item

    def delete_item(self, item_index):
        """Delete a HeapNode in a specific position (index) within heap_list"""
        # locate target node and replace with last node in the heap_list
        target = self.heap_list[item_index]
        last_item = self.heap_list.pop()
        if item_index == self.heap_size:
            self.heap_size -= 1
        else:
            self.heap_list[item_index] = last_item
            self.heap_size -= 1
            self._percolate_down(item_index)  # percolate down to place replacement node in the correct place
        return target

    def build(self, new_list):
        """Construct a MaxHeap given a list of numbers or tuples"""
        for i in new_list:
            self.insert(i)

    def heap_change_key(self, val, new_key):
        """For heaps containing HeapNodes of tuples - change the keys of items with values val"""
        for index, item in enumerate(self.heap_list):
            if index > 0:
                # delete the node with matching val
                if item.get_data()[1] == val:
                    self.delete_item(index)
        # insert a new node object with the new key and same val
        self.insert((new_key, val))

    def __repr__(self):
        return str([node for node in self.heap_list if node.get_data()])

    def __contains__(self, val):
        contained = False
        for node in self.heap_list:
            if type(node.get_data()) is tuple:
                if node.get_data()[1] == val:
                    contained = True
                    break
            else:
                if node.get_data() == val:
                    contained = True
                    break
        return contained

    def __iter__(self):
        return iter(node.get_data() for node in self.heap_list)


if __name__ == '__main__':
    bin_heap = MaxHeap()
    # new_list = [HeapNode((9, 'dog')), HeapNode((5, 'cat')), HeapNode((6, 'bear')), HeapNode((2, 'horse')),
    #             HeapNode((3, 'snake')), HeapNode((4, 'ant')), HeapNode((8, 'bird'))]
    new_list = [(9, 'dog'), (5, 'cat'), (6, 'bear'), (2, 'horse'), (3, 'snake'), (4, 'ant'), (8, 'bird')]
    bin_heap.build(new_list)
    print(bin_heap)
    print(bin_heap)
    print(bin_heap.pop())
    print(bin_heap.pop())
    bin_heap.insert((1, 'Shark'))
    bin_heap.delete_item(3)
    print(bin_heap.heap_list)
    bin_heap.heap_change_key('cat', 10)
    print(bin_heap)
    print('lion' in bin_heap)



from PyDSL.binary_tree import BinaryTree
from PyDSL.queue import Queue


def bfs(tree):
    """Perform breadth-first search of the tree.

    Parameters:
        tree: the tree to perform BFS on

    Returns:
        A list of nodes in order of the search
    """
    bfs_list = []
    node_queue = Queue()
    node_queue.enqueue(tree.root)
    while not node_queue.is_empty():
        # get the node first-in first-out
        current_node = node_queue.dequeue()
        # get the children and place into queue
        if current_node.get_left_child():
            node_queue.enqueue(current_node.get_left_child())
        if current_node.get_right_child():
            node_queue.enqueue(current_node.get_right_child())
        # attach node to list
        bfs_list.append(current_node)

    return bfs_list


def _preorder(tree, current_node=None, order_list=[], root_set=False):
    """Helper method to implement preorder traversal of the tree"""
    # find the root and set as current_node for first iteration
    if not current_node and not root_set:
        root_set = True
        current_node = tree.root
    if current_node:
        # search the tree recursively and append the nodes to order_list in preorder fashion
        order_list.append(current_node)
        _preorder(tree, current_node=current_node.get_left_child(), root_set=root_set)
        _preorder(tree, current_node=current_node.get_right_child(), root_set=root_set)

    return order_list


def preorder(tree):
    """Perform preorder tree traversal.

        Parameters:
            tree: the tree to perform preorder traversal on

        Returns:
            A list of nodes in order of the search
    """
    return _preorder(tree)


def _inorder(tree, current_node=None, order_list=[], root_set=False):
    """Helper method to implement inorder traversal of the tree"""
    if not current_node and not root_set:
        # find the root and set as current_node for first iteration
        root_set = True
        current_node = tree.root
    if current_node:
        # search the tree recursively and append the nodes to order_list in inorder fashion
        _inorder(tree, current_node=current_node.get_left_child(), root_set=root_set)
        order_list.append(current_node)
        _inorder(tree, current_node=current_node.get_right_child(), root_set=root_set)

    return order_list


def inorder(tree):
    """Perform inorder tree traversal.

        Parameters:
            tree: the tree to perform inorder traversal on

        Returns:
            A list of nodes in order of the search
    """
    return _inorder(tree)


def _postorder(tree, current_node=None, order_list=[], root_set=False):
    """Helper method to implement postorder traversal of the tree"""
    if not current_node and not root_set:
        # find the root and set as current_node for first iteration
        root_set = True
        current_node = tree.root
    if current_node:
        # search the tree recursively and append the nodes to order_list in postorder fashion
        _postorder(tree, current_node=current_node.get_left_child(), root_set=root_set)
        _postorder(tree, current_node=current_node.get_right_child(), root_set=root_set)
        order_list.append(current_node)

    return order_list


def postorder(tree):
    """Perform postorder tree traversal.

        Parameters:
            tree: the tree to perform postorder traversal on

        Returns:
            A list of nodes in order of the search
    """
    return _postorder(tree)


if __name__ == '__main__':
    tree = BinaryTree()
    key_list = [43, 55, 22, 15, 98, 1, 34, 76]
    val_list = ['Penguin', 'Ant', 'Bear', 'Shark', 'Eagle', 'Dog', 'Cat', 'Turtle']

    for i, j in zip(key_list, val_list):
        tree[i] = j

    print(tree[43].get_val())
    print(43 in tree)

    print([i.get_val() for i in preorder(tree)])
    print([i.get_val() for i in inorder(tree)])
    print([i.get_val() for i in postorder(tree)])
    print([i.get_val() for i in bfs(tree)])










class Node:
    """Node class for graphs

        Attributes:
            key: key of the node
            connections: a dictionary containing the neighbours of the node and the weight of the edge as a key-val pair
            colour: flag for search algorithms. Default = 'white'
    """

    def __init__(self, key):
        """Initialise the graph node object"""
        self.key = key
        self.connections = {}
        self.colour = 'white'

    def add_nbr(self, nbr, weight):
        """Add a neighbour, and specify the weight of the edge"""
        self.connections[nbr] = weight

    def get_nbrs(self):
        """Get the keys of all the neighbours in a list"""
        return self.connections.keys()

    def get_key(self):
        """Get the key of the node"""
        return self.key

    def get_weight(self, nbr):
        """Get the weight of the edge connecting the node and a neighbour"""
        return self.connections[nbr]

    def get_colour(self):
        """Get the colour of the node"""
        return self.colour

    def set_colour(self, col):
        """Set the colour of the node"""
        self.colour = col

    def __repr__(self):
        return 'Node {}: contains connections to {}'.format(self.key, self.connections)


class Graph:
    """Graph class

        Attributes:
            node_list: a list of all nodes belonging to the graph
            size: the number of nodes in the graph
    """

    def __init__(self):
        """Initialise an empty graph"""
        self.node_list = {}
        self.size = 0

    def add_node(self, key):
        """Add a node to the graph"""
        self.size += 1
        new_node = Node(key)
        self.node_list[key] = new_node
        return new_node

    def get_node(self, key):
        """Get the node in the graph with corresponding key"""
        if key in self.node_list:
            return self.node_list[key]
        else:
            return None

    def add_edge(self, key, edge, weight):
        """Add an edge between two nodes (from key to edge) with specified weight"""
        target = self.get_node(key)
        if target:
            # add edge from Node(key) to Node(edge)
            target.add_nbr(edge, weight)
        else:
            return 'Cannot add edge. Selected node does not exist'

    def add_undirected_edge(self, key, edge, weight):
        """Add an undirected edge between two nodes (from key to edge) with specified weight"""
        target = self.get_node(key)
        targeted = self.get_node(edge)
        if target and targeted:
            # add edge from Node(key) to Node(edge)
            target.add_nbr(edge, weight)
            # add edge from Node(edge) to Node(key)
            targeted.add_nbr(key, weight)
        else:
            return 'Cannot add undirected edge. One of the selected nodes does not exist'

    def get_nodes(self):
        """Get the keys of all the nodes in the graph"""
        return self.node_list.keys()

    def __contains__(self, key):
        return key in self.node_list

    def __iter__(self):
        return iter(self.node_list.values())















from PyDSL.graph import Graph
from PyDSL.graph import Node
from PyDSL.priority_queue import MinPriorityQueue
from PyDSL.queue import Queue
from PyDSL.stack import Stack


class DistanceNode(Node):
    """Node class for graph algorithms requiring distance and predecessor attributes. Extends the graph node class

        Attributes:
            key: key of the node
            connections: a dictionary containing the neighbours of the node and the weight of the edge as a key-val pair
            colour: flag for search algorithms. Default = 'white'
            distance: the distance attribute. Default = 0
            pred: the predecessor node. Default = None
        """

    def __init__(self, key):
        """Initialise the graph DistanceNode object

            Parameters:
                key: the key of the DistanceNode
            """
        super().__init__(key)
        self.distance = 0
        self.pred = None

    def set_distance(self, dist):
        """Set the distance of the node"""
        self.distance = dist

    def get_distance(self):
        """Get the distance of the node"""
        return self.distance

    def set_pred(self, node):
        """Set the predecessor node of the node"""
        self.pred = node

    def get_pred(self):
        """Get the predecessor node of the node"""
        return self.pred


def dijkstra(graph, start_node):
    """Perform Dijkstra's Algorithm on the graph, given a starting node.

        Parameters:
            graph: the graph to perform Dijkstra on
            start_node: the starting node, relative to which distances are found

        Returns:
            Returns a graph of DistanceNode objects, with the predecessor of each node corresponding
            to the shortest path from said node to the starting node
    """

    # initialise an empty graph, to replicate input graph but with DistanceNodes instead of Nodes
    d_graph = Graph()
    d_start_node = None
    for node in graph:
        # create DistanceNode
        d_node = DistanceNode(node.get_key())
        for nbr in node.get_nbrs():
            # connect new DistanceNode to all its neighbours
            nbr_weight = node.get_weight(nbr)
            d_node.add_nbr(nbr, nbr_weight)
        # set the node distance to infinity
        d_node.set_distance(float('inf'))
        # place DistanceNode into new graph
        d_graph.node_list[d_node.get_key()] = d_node
        # set the starting DistanceNode
        if node == start_node:
            d_start_node = d_node

    pq = MinPriorityQueue()
    # set the start node distance to 0 and enqueue all the nodes in the graph
    d_start_node.set_distance(0)
    # place tuple of (node distance, node) into priority queue - this will prioritise based on node distance
    pq.enqueue_list([(node.get_distance(), node) for node in d_graph])
    while pq.get_size() > 0:
        current_node = pq.dequeue().get_data()[1]
        # if the current_node has distance = inf, the node is in a disconnected portion of the graph. Stop searching
        if current_node.get_distance() == float('inf'):
            break
        for nbr in current_node.get_nbrs():
            # look through neighbours of current_node,
            # and set new distance = current_node distance + edge weight
            nbr_node = d_graph.get_node(nbr)
            new_dist = current_node.get_distance() + current_node.get_weight(nbr)
            if new_dist < nbr_node.get_distance():
                # if the new distance is less than the original neighbour node distance, set the current_node as
                # its predecessor and set its distance as new_dist.
                nbr_node.set_distance(new_dist)
                nbr_node.set_pred(current_node)
                # change the key of the node in the priority queue to reflect the new distance
                pq.change_key(nbr_node, new_dist)

    return d_graph


def prim(graph):
    """Performs Prim's Algorithm on the graph.

        Parameters:
            graph: the graph to find the MST of a graph

        Returns:
            A list of minimum spanning trees (graphs)
    """

    # initialise an empty graph, to replicate input graph but with DistanceNodes instead of Nodes
    d_graph = Graph()
    span_tree = None
    span_forest = []

    for node in graph:
        # convert nodes in the graph to DistanceNodes
        d_node = DistanceNode(node.get_key())
        for nbr in node.get_nbrs():
            nbr_weight = node.get_weight(nbr)
            d_node.add_nbr(nbr, nbr_weight)
        # set the distances of all nodes to inf
        d_node.set_distance(float('inf'))
        # put node into the new graph
        d_graph.node_list[d_node.get_key()] = d_node

    pq = MinPriorityQueue()
    # place tuple of (node distance, node) into priority queue - this will prioritise based on node distance
    pq.enqueue_list([(node.get_distance(), node) for node in d_graph])
    last_node = None
    while pq.get_size() > 0:
        current_node = pq.dequeue().get_data()[1]
        if last_node and current_node.get_distance() != float('inf'):
            # add the new node to the spanning tree and connect to the last node
            span_tree.add_node(current_node.get_key())
            span_tree.add_undirected_edge(last_node.get_key(), current_node.get_key(),
                                          last_node.get_weight(current_node.get_key()))
        elif current_node.get_distance() == float('inf'):
            # if the current node has distance = inf, start a new spanning tree and set dist = 0
            if span_tree:
                span_forest.append(span_tree)
            span_tree = Graph()
            current_node.set_distance(0)
            span_tree.add_node(current_node.get_key())
        else:
            # if not last node and dist != inf, just add the node
            span_tree.add_node(current_node.get_key())
        for nbr in current_node.get_nbrs():
            # get the node and the edge weight to each neighbour
            nbr_node = d_graph.get_node(nbr)
            new_dist = current_node.get_weight(nbr)
            if nbr_node in pq and new_dist < nbr_node.get_distance():
                # set the neighbour distance as the weight between current_node and neighbour
                nbr_node.set_distance(new_dist)
                # change the key of the node in the priority queue to reflect the new distance
                pq.change_key(nbr_node, new_dist)
        last_node = current_node

    span_forest.append(span_tree)

    return span_forest


def _bfs_visit(graph, current_node):
    """Helper method for graph breadth-first search"""
    q = Queue()
    g = Graph()  # initialise the spanning tree
    q.enqueue(current_node)
    while q.get_size() > 0:
        current_node = q.dequeue()
        # add the new node to the spanning tree
        g.add_node(current_node.get_key())
        # set the colour to black after adding to tree
        current_node.set_colour('black')
        for nbr in current_node.get_nbrs():
            # get the neighbour nodes
            nbr_node = graph.get_node(nbr)
            if nbr_node.get_colour() == 'white':
                # if the neighbour node is unexplored (white), set to waiting (grey)
                nbr_node.set_colour('grey')
                # add edge between current node and neighbour. Then put neighbour in queue
                g.add_edge(current_node.get_key(), nbr_node.get_key(), current_node.get_weight(nbr))
                q.enqueue(nbr_node)

    return g


def bfs(graph):
    """Perform breadth-first search on the graph.

        Parameters:
            graph: the graph to perform BFS on

        Returns:
            Returns a list of spanning trees (graphs)
    """

    tree_list = []  # initialise forest (list of trees)
    # search through graph and call bfs on all unexplored nodes. Append to forest
    for node in graph:
        if node.get_colour() == 'white':
            spanning_tree = _bfs_visit(graph, node)
            tree_list.append(spanning_tree)
    return tree_list


def _dfs_visit(graph, current_node):
    """Helper method for graph depth-first search"""
    s = Stack()
    g = Graph()  # initialise the spanning tree
    s.push(current_node)
    current_node.set_colour('grey')
    # add the current_node to the spanning tree
    g.add_node(current_node.get_key())
    while s.get_size() > 0:
        current_node = s.peek()
        all_done = True  # flag for all neighbours in waiting
        for nbr in current_node.get_nbrs():
            nbr_node = graph.get_node(nbr)
            if nbr_node.get_colour() == 'white':
                all_done = False  # if neighbour is unexplored, set flag to false
                s.push(nbr_node)
                nbr_node.set_colour('grey')
                # add neighbour node to the graph and connect current_node and the neighbour
                g.add_node(nbr_node.get_key())
                g.add_edge(current_node.get_key(), nbr_node.get_key(), current_node.get_weight(nbr))
                break  # escape from the current node, select a new node from the stack
        if all_done:
            # remove the node from stack once all neighbours are in waiting
            current_node.set_colour('black')
            s.pop()

    return g


def dfs(graph):
    """Perform depth-first search on the graph.

           Parameters:
               graph: the graph to perform DFS on

           Returns:
               Returns a list of spanning trees (graphs)
    """
    tree_list = []  # initialise forest (list of trees)
    for node in graph:
        # search through graph and call dfs on all unexplored nodes. Append to forest
        if node.get_colour() == 'white':
            spanning_tree = _dfs_visit(graph, node)
            tree_list.append(spanning_tree)
    return tree_list


def _top_dfs_visit(graph, current_node):
    """Helper method for topological sort"""
    order_list = []
    s = Stack()
    s.push(current_node)
    current_node.set_colour('grey')
    while s.get_size() > 0:
        current_node = s.peek()
        all_done = True  # flag for all neighbours in waiting
        for nbr in current_node.get_nbrs():
            nbr_node = graph.get_node(nbr)
            if nbr_node.get_colour() == 'white':
                all_done = False  # if one of the neighbours is unexplored, set flag as false
                # add the neighbour to the stack and then set neighbour to in waiting
                s.push(nbr_node)
                nbr_node.set_colour('grey')
                break  # escape from the current node, select a new node from the stack
        if all_done:
            # remove the node from stack once all neighbours are in waiting
            current_node.set_colour('black')
            s.pop()
            order_list.append(current_node)  # place the node in the order_list

    return order_list


def topological_sort(graph):
    """Performs topological sort on a directed acyclic graph.

        Parameters:
            graph: the graph to perform topological sort on

        Returns:
            A list of sorted nodes
    """

    top_list = []
    for node in graph:
        # search through graph and call topological sort on all unexplored nodes
        if node.get_colour() == 'white':
            order_list = _top_dfs_visit(graph, node)
            # combine the lists
            top_list = top_list + order_list

    return [i for i in reversed(top_list)]


if __name__ == '__main__':
    g = Graph()
#         g.add_node('u')
#         g.add_node('x')
#         g.add_node('v')
#         g.add_node('w')
#         g.add_node('y')
#         g.add_node('z')
#         g.add_node('a')
#         g.add_node('b')
#
#         g.add_edge('u', 'v', 2)
#         g.add_edge('u', 'x', 1)
#         g.add_edge('u', 'w', 5)
#         g.add_edge('w', 'u', 5)
#         g.add_edge('x', 'u', 1)
#         g.add_edge('v', 'u', 2)
#
#         g.add_edge('x', 'v', 2)
#         g.add_edge('v', 'x', 2)
#         g.add_edge('x', 'w', 3)
#         g.add_edge('w', 'x', 3)
#         g.add_edge('v', 'w', 3)
#         g.add_edge('w', 'v', 3)
#
#         g.add_edge('x', 'y', 1)
#         g.add_edge('y', 'x', 1)
#         g.add_edge('y', 'w', 1)
#         g.add_edge('w', 'y', 1)
#         g.add_edge('y', 'z', 1)
#         g.add_edge('z', 'y', 1)
#
#         g.add_edge('w', 'z', 5)
#         g.add_edge('z', 'w', 5)
#
#         g.add_edge('a', 'b', 2)
#
#         start_node = g.get_node('u')
#
#         d_g = dijkstra(g, start_node)
#         print(d_g)
#
#         dest_node = d_g.get_node('w')
#         x = dest_node
#         print(x.key)
#         while x.pred:
#             print(x.pred.key)
#             x = x.pred


    # g = Graph()
    g.add_node('A')
    g.add_node('B')
    g.add_node('C')
    g.add_node('D')
    g.add_node('E')
    g.add_node('F')
    g.add_node('G')

    g.add_undirected_edge('A', 'E', 1)
    g.add_undirected_edge('A', 'B', 3)
    g.add_undirected_edge('B', 'C', 9)
    g.add_undirected_edge('B', 'D', 2)
    g.add_undirected_edge('B', 'E', 2)
    g.add_undirected_edge('C', 'D', 3)
    g.add_undirected_edge('C', 'E', 7)
    g.add_undirected_edge('F', 'G', 6)

    minspan = prim(g)
    print(minspan)
    for i in minspan:
        for j in i:
            print(j)


    # g.add_node('a')
    # g.add_node('b')
    # g.add_node('c')
    # g.add_node('d')
    # g.add_node('e')
    # g.add_node('f')
    # g.add_node('g')
    # g.add_node('h')
    # g.add_node('i')
    # g.add_node('j')
    # g.add_node('k')
    #
    # g.add_undirected_edge('a', 'c', 3)
    # g.add_undirected_edge('b', 'c', 0)
    # g.add_undirected_edge('c', 'e', 4)
    # g.add_undirected_edge('e', 'd', 0)
    # g.add_undirected_edge('d', 'f', 1)
    # g.add_undirected_edge('e', 'f', 9)
    # g.add_undirected_edge('f', 'g', 3)
    # g.add_undirected_edge('g', 'h', 0)
    # g.add_undirected_edge('f', 'h', 7)
    # g.add_undirected_edge('h', 'i', 6)
    # g.add_undirected_edge('h', 'k', 3)
    # g.add_undirected_edge('k', 'j', 0)
    #
    # spantree = dfs(g)
    # for i in spantree:
    #     for j in i:
    #         print(j)



from PyDSL.linked_list import Node
from PyDSL.linked_list import LinkedList


class Stack(LinkedList):
    """Stack class (last-in first-out). Extends LinkedList

        Attributes:
            head: Node object. The head of the linked list
            size: the length of the linked list
    """

    def __init__(self):
        """Initialise the stack object"""
        super().__init__()

    def push(self, item):
        """Add item to the stack"""
        item_to_add = Node(item)
        item_to_add.set_next(self.head)  # connect new node to old head
        self.head = item_to_add  # set new node as new head
        self.size += 1

    def pop(self):
        """Remove and return item"""
        first_item = self.head
        next_item = self.head.get_next()  # set new head
        self.head = next_item
        self.size -= 1

        # return the original head
        return first_item.get_data()

    def peek(self):
        """Return the next item from the stack"""
        return self.head.get_data()





"""
Linked list class
"""


class Node:
    """Node class for linked list

        Attributes:
            data: the data contained inside the node
            next: Node object. The next node in the linked list
            previous: Node object. The previous node in the linked list
    """

    def __init__(self, init_data):
        """Initialise the TreeNode object

            Arguments:
                init_data: the data of the node
        """
        self.data = init_data
        self.next = None
        self.previous = None

    def get_data(self):
        """Get data of node"""
        return self.data

    def get_next(self):
        """Get the next node"""
        return self.next

    def get_prev(self):
        """Get the previous node"""
        return self.previous

    def set_data(self, input_data):
        """Set data for the node"""
        self.data = input_data

    def set_next(self, input_next):
        """Set the next node"""
        self.next = input_next

    def set_prev(self, input_prev):
        """Set the previous node"""
        self.previous = input_prev


class LinkedList:
    """Node class for linked list

        Attributes:
            head: Node object. The head of the linked list
            size: the length of the linked list
    """

    def __init__(self):
        """Initialise the TreeNode object"""
        self.head = None
        self.size = 0

    def is_empty(self):
        """Check if the linked list is empty"""
        if self.size == 0:
            return True
        else:
            return False

    def remove(self, item):
        """Remove a node with matching data"""
        current = self.head
        previous = None
        found = False
        # follow the linked list until the correct node is found
        while not current and not found:
            if current.get_data() == item:
                found = True
            else:
                previous = current
                current = current.get_next()
        # remove the correct node and then re-link the linked list
        if found and not previous:
            self.head = current.get_next()
        elif found and previous:
            previous.set_next(current.get_next())
        else:
            raise KeyError('Item not in linked list')

    def get_size(self):
        """Get the size of the LinkedList"""
        return self.size

    def search(self, item):
        """Search for a node with matching data. Return True if found, False if not"""
        current = self.head
        found = False
        # follow the linked list until the correct node is found
        while not current and not found:
            if current.get_data() == item:
                found = True
            else:
                current = current.get_next()

        if not current and not found:
            return False
        else:
            return True

    def insert(self, item, index):
        """Insert a node at an index"""
        item_to_add = Node(item)
        current = self.head
        previous = None
        count = 0
        # follow the linked list to the correct index
        while count < index and current:
            previous = current
            current = current.get_next()
            count += 1
        # perform insertion
        if previous:
            if current:
                item_to_add.set_next(current)
                previous.set_next(item_to_add)
            else:
                print('Cannot insert - index out of range')
        else:
            item_to_add.set_next(current)
            self.head = item_to_add

    












#! usr/bin/env python2.7

from PyDSL.heap import MaxHeap
from PyDSL.heap import MinHeap


class MaxPriorityQueue(MaxHeap):
    """Heap-based max priority queue. The node with maximum key is prioritised.
        Provides queue-like methods for the MaxHeap class. Extends the MaxHeap class

        Attributes:
            heap_list: the list of HeapNodes. The first element is always an empty node that is ignored by operations
            heap_size: the number of elements in the heap. Starts at 0
    """

    def __init__(self):
        """Initialise the MaxPriorityQueue"""
        super().__init__()

    def enqueue(self, item):
        """Insert item into MaxPriorityQueue"""
        self.insert(item)

    def dequeue(self):
        """Remove item from MaxPriorityQueue"""
        return self.pop()

    def is_empty(self):
        """Check if MaxPriorityQueue is empty"""
        return self.heap_size == 0

    def peek(self):
        """Return the first element of the MaxPriorityQueue"""
        return self.heap_list[1]

    def get_size(self):
        """Get the size of the MaxPriorityQueue"""
        return self.heap_size

    def enqueue_list(self, en_list):
        """Converts a list into a MaxPriorityQueue"""
        self.build(en_list)

    def change_key(self, val, new_key):
        """For priority queues containing tuples - change the keys of items with values val"""
        self.heap_change_key(val, new_key)

    def __repr__(self):
        return '{}'.format([item for item in self.heap_list if item != 0])


class MinPriorityQueue(MinHeap):
    """Heap-based min priority queue. The node with minimum key is prioritised.
        Provides queue-like methods for the MinHeap class. Extends the MinHeap class

        Attributes:
            heap_list: the list of HeapNodes. The first element is always an empty node that is ignored by operations
            heap_size: the number of elements in the heap. Starts at 0
    """

    def __init__(self):
        """Initialise the MinPriorityQueue"""
        super().__init__()

    def enqueue(self, item):
        """Insert item into MinPriorityQueue"""
        self.insert(item)

    def dequeue(self):
        """Remove item from MinPriorityQueue"""
        return self.pop()

    def is_empty(self):
        """Check if MinPriorityQueue is empty"""
        return self.heap_size == 0

    def peek(self):
        """Return the first element of the MinPriorityQueue"""
        return self.heap_list[1]

    def get_size(self):
        """Get the size of the MinPriorityQueue"""
        return self.heap_size

    def enqueue_list(self, en_list):
        """Converts a list into a MinPriorityQueue"""
        self.build(en_list)

    def change_key(self, val, new_key):
        """For priority queues containing tuples - change the keys of items with values val"""
        self.heap_change_key(val, new_key)

    def __repr__(self):
        return '{}'.format([item for item in self.heap_list if item != 0])


if __name__ == '__main__':
    pq = MinPriorityQueue()
    new_list = [(9, 'dog'), (5, 'cat'), (6, 'bear'), (2, 'horse'), (3, 'snake'), (4, 'ant'), (8, 'bird')]
    pq.enqueue_list(new_list)
    print(pq.peek())
    print(pq.dequeue())
    print(pq.get_size())
    print(pq.dequeue())
    print(pq.get_size())
    print(pq.dequeue())
    print(pq.dequeue())
    print(pq)








from PyDSL.linked_list import Node
from PyDSL.linked_list import LinkedList


class Deque(LinkedList):
    """Node class for deque. Extends LinkedList

        Attributes:
            head: Node object. The head of the deque
            tail: Node object. The tail of the deque
            size: the length of the deque
    """

    def __init__(self):
        """Initialise Deque class"""
        super().__init__()
        self.tail = None

    def left_enqueue(self, item):
        """Add an item to the left of the deque"""
        item_to_add = Node(item)
        item_to_add.set_next(self.head)

        # if the deque is empty, the new item is the tail
        if not self.tail:
            self.tail = item_to_add
        else:
            # connect the old head to the new head
            self.head.set_prev(item_to_add)

        # set the new node as the head
        self.head = item_to_add
        self.size += 1

    def right_enqueue(self, item):
        """Add an item to the right of the deque"""
        item_to_add = Node(item)
        item_to_add.set_prev(self.tail)

        # if the deque is empty, the new item is the head
        if not self.head:
            self.head = item_to_add
        else:
            # connect the old tail to the new tail
            self.tail.set_next(item_to_add)

        self.tail = item_to_add
        self.size += 1

    def left_dequeue(self):
        """Remove an item from the left of the deque and return it"""
        # get the next head and set it as the new head. Save the old head
        next_head = self.head.get_next()
        head = self.head
        self.head = next_head
        self.size -= 1

        # return the data in the old head
        return head.get_data()

    def right_dequeue(self):
        """Remove an item from the right of the deque and return it"""
        # get the next tail and set it as the new tail. Save the oldtail
        next_last = self.tail.get_prev()
        last = self.tail
        self.tail = next_last
        self.size -= 1

        # return the data in the old tail
        return last.get_data()







from PyDSL.AVL_tree import AVLTree
from random import randint
from typing import List, Tuple, Callable

# TREE_LAYER_DIFF = 3


def __print_branch(branch, layer_height: int, print_fn: Callable) -> Tuple[int, int, List[str]]:
    if branch is None:
        return 0, 0, []

    val_string = print_fn(branch)

    if branch.left_child is None and branch.right_child is None:
        return len(val_string), int(len(val_string) / 2), [val_string]

    left_branch_width, left_tree_centre, left_branch_lines = __print_branch(branch.left_child, layer_height, print_fn)
    right_branch_width, right_tree_centre, right_branch_lines = __print_branch(branch.right_child, layer_height, print_fn)

    central_padding = max(1, len(val_string) - left_branch_width - right_branch_width)

    line_width = left_branch_width + central_padding + right_branch_width

    small_side = min(len(left_branch_lines), len(right_branch_lines))
    large_side = max(len(left_branch_lines), len(right_branch_lines))

    centre = int(left_branch_width + central_padding / 2)

    left_element_padding = max(0, centre - round(len(val_string) / 2))
    right_element_padding = line_width - len(val_string) - left_element_padding

    if right_element_padding < 0:
        left_element_padding += right_element_padding
        right_element_padding = 0

    lines = [" " * left_element_padding + val_string + " " * right_element_padding]

    sx = centre
    dxl = left_tree_centre - sx
    dxr = (left_branch_width + central_padding + right_tree_centre) - sx

    for i in range(layer_height):
        line_str = [" " for _ in range(line_width)]

        if left_branch_width != 0:
            pos = sx + (i / (layer_height - 1)) * dxl

            # if i < layer_height - 1:
            #     for fill in range(round(pos), round(pos + dxl / (2 * layer_height)), -1):
            #         line_str[fill] = "-"
            # if i > 0:
            #     for fill in range(round(pos), round(pos - dxl / (2 * layer_height)), 1):
            #         line_str[fill] = "-"
            line_str[round(pos)] = "/"

        if right_branch_width != 0:
            pos = sx + (i / (layer_height - 1)) * dxr
            line_str[round(pos)] = "\\"

        lines.append("".join(line_str))

    for i in range(0, small_side):
        lines.append(left_branch_lines[i] + " " * central_padding + right_branch_lines[i])

    if len(left_branch_lines) > len(right_branch_lines):
        for i in range(small_side, large_side):
            lines.append(left_branch_lines[i] + " " * (right_branch_width + central_padding))
    else:
        for i in range(small_side, large_side):
            lines.append(" " * (left_branch_width + central_padding) + right_branch_lines[i])

    return line_width, centre, lines


def pretty_print_tree(tree, layer_height=2, print_fn=lambda br: f"{{{br.key}:{br.val}}}"):
    _, _, lines = __print_branch(tree.root, layer_height, print_fn)
    for line in lines:
        print(line)


def pretty_print_graph(graph):
    pass


if __name__ == '__main__':
    tree = AVLTree()
    for i in range(0, 50):
        P = randint(0, 1000)
        tree.insert_node(P, i)
    pretty_print_tree(tree, 6)







from PyDSL.linked_list import Node
from PyDSL.linked_list import LinkedList


class Queue(LinkedList):
    """Queue class (first-in first-out). Extends LinkedList

        Attributes:
            head: Node object. The head of the linked list
            size: the length of the linked list
            tail: Node object. The tail of the linked list
    """

    def __init__(self):
        """Initialise the queue"""
        super().__init__()
        self.tail = None

    def enqueue(self, item):
        """Insert item into the queue"""
        item_to_add = Node(item)
        item_to_add.set_next(self.head)

        # if the queue is empty, the new item is the tail
        if not self.tail:
            self.tail = item_to_add
        else:
            # connect the old head to the new head
            self.head.set_prev(item_to_add)

        self.head = item_to_add
        self.size += 1

    def dequeue(self):
        """Remove and return item from the queue"""
        # get the next head and set it as the new head. Save the old head
        next_last = self.tail.get_prev()
        last = self.tail
        self.tail = next_last
        self.size -= 1

        return last.get_data()

    def peek(self):
        """Return the next item from the queue"""
        return self.tail.get_data()





